"""The DniDatetime class."""

from datetime import datetime, timedelta, timezone
from typing import Any, Optional, TYPE_CHECKING, Union

from .constants import (
    DniDateTimeTuple,
    DNI_EPOCH_HAHR,
    DNI_UNITS,
    MAX_DNI_ORDINAL
)
from .conversions import dni_to_earth, earth_to_dni, hvy2ord
from .dnidate import DniDate
from .dnitime import DniTime
from .dnitimedelta import DniTimedelta
from .utils import check_dni_date_fields, check_dni_time_fields, cmp


class DniDatetime:
    """A D'ni timestamp."""

    __slots__ = DniDate.__slots__ + DniTime.__slots__

    if TYPE_CHECKING:
        _hahr: int
        _vailee: int
        _yahr: int
        _gahrtahvo: int
        _tahvo: int
        _gorahn: int
        _prorahn: int

        # Class properties
        min: "DniDatetime"
        max: "DniDatetime"
        resolution: DniTimedelta

    def __new__(
        cls,
        hahr: int = DNI_UNITS["hahr"]["min"],
        vailee: int = DNI_UNITS["vailee"]["min"],
        yahr: int = DNI_UNITS["yahr"]["min"],
        gahrtahvo: int = DNI_UNITS["gahrtahvo"]["min"],
        tahvo: int = DNI_UNITS["tahvo"]["min"],
        gorahn: int = DNI_UNITS["gorahn"]["min"],
        prorahn: int = DNI_UNITS["prorahn"]["min"],
    ) -> "DniDatetime":
        if isinstance(hahr, int):
            hahr, vailee, yahr = check_dni_date_fields(hahr, vailee, yahr)
            gahrtahvo, tahvo, gorahn, prorahn = check_dni_time_fields(
                gahrtahvo, tahvo, gorahn, prorahn
            )
            self = object.__new__(cls)
            self._hahr = hahr
            self._vailee = vailee
            self._yahr = yahr
            self._gahrtahvo = gahrtahvo
            self._tahvo = tahvo
            self._gorahn = gorahn
            self._prorahn = prorahn
            return self
        return NotImplemented

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}"
            f"({self._hahr}, "
            f"{self._vailee}, "
            f"{self._yahr}, "
            f"{self._gahrtahvo}, "
            f"{self._tahvo}, "
            f"{self._gorahn}, "
            f"{self._prorahn})"
        )

    def isoformat(self) -> str:
        """Return the datetime formatted according to ISO standard."""
        return (
            f"{self._hahr}-{self._vailee:02}-{self._yahr:02} "
            f"{self._gahrtahvo}:{self._tahvo:02}:"
            f"{self._gorahn:02}:{self._prorahn:02}"
        )

    __str__ = isoformat

    # Read-only field accessors
    @property
    def hahr(self) -> int:
        """hahr"""
        return self._hahr

    @property
    def vailee(self) -> int:
        """vailee"""
        return self._vailee

    @property
    def yahr(self) -> int:
        """yahr"""
        return self._yahr

    @property
    def gahrtahvo(self) -> int:
        """gahrtahvo"""
        return self._gahrtahvo

    @property
    def tahvo(self) -> int:
        """tahvo"""
        return self._tahvo

    @property
    def gorahn(self) -> int:
        """gorahn"""
        return self._gorahn

    @property
    def prorahn(self) -> int:
        """prorahn"""
        return self._prorahn

    def replace(
        self,
        hahr: Optional[int] = None,
        vailee: Optional[int] = None,
        yahr: Optional[int] = None,
        gahrtahvo: Optional[int] = None,
        tahvo: Optional[int] = None,
        gorahn: Optional[int] = None,
        prorahn: Optional[int] = None,
    ) -> "DniDatetime":
        """Return a new D'ni datetime with new values for the specified fields."""
        if hahr is None:
            hahr = self._hahr
        if vailee is None:
            vailee = self._vailee
        if yahr is None:
            yahr = self._yahr
        if gahrtahvo is None:
            gahrtahvo = self._gahrtahvo
        if tahvo is None:
            tahvo = self._tahvo
        if gorahn is None:
            gorahn = self._gorahn
        if prorahn is None:
            prorahn = self._prorahn
        return type(self)(hahr, vailee, yahr, gahrtahvo, tahvo, gorahn, prorahn)

    # Arithmetic
    def __add__(self, other: Any) -> "DniDatetime":
        if not isinstance(other, (timedelta, DniTimedelta)):
            return NotImplemented
        if isinstance(other, timedelta):
            other = DniTimedelta.from_timedelta(other)
        delta = DniTimedelta(
            self.to_ordinal(),
            gahrtahvotee=self._gahrtahvo,
            tahvotee=self._tahvo,
            gorahntee=self._gorahn,
            prorahntee=self._prorahn,
        )
        delta += other
        total_prorahn = delta.pahrtahvotee * 3125 + delta.prorahntee
        gahrtahvo, rem = divmod(total_prorahn, 15625)
        tahvo, rem = divmod(rem, 625)
        gorahn, prorahn = divmod(rem, 25)
        if 0 < delta.yahrtee <= MAX_DNI_ORDINAL:
            return type(self).combine(
                DniDate.from_ordinal(delta.yahrtee),
                DniTime(gahrtahvo, tahvo, gorahn, prorahn),
            )
        raise OverflowError("result out of range")

    __radd__ = __add__

    def __sub__(self, other: Any) -> Union["DniDatetime", DniTimedelta]:
        if not isinstance(other, DniDatetime):
            if isinstance(other, timedelta):
                return self + -other
        else:
            yahrtee1 = self.to_ordinal()
            yahrtee2 = other.to_ordinal()
            prorahntee1 = (
                self._gahrtahvo * 15625
                + self._tahvo * 625
                + self._gorahn * 25
                + self._prorahn
            )
            prorahntee2 = (
                other._gahrtahvo * 15625
                + other._tahvo * 625
                + other._gorahn * 25
                + other._prorahn
            )
            return DniTimedelta(yahrtee1 - yahrtee2, prorahntee1 - prorahntee2)
        return NotImplemented

    # Rich comparisons
    def _cmp(self, other: Any) -> int:
        """Helper function for rich comparison of DniDatetime objects."""
        # pylint: disable=protected-access
        assert isinstance(other, DniDatetime)
        return cmp(
            (
                self._hahr,
                self._vailee,
                self._yahr,
                self._gahrtahvo,
                self._tahvo,
                self._gorahn,
                self._prorahn,
            ),
            (
                other._hahr,
                other._vailee,
                other._yahr,
                other._gahrtahvo,
                other._tahvo,
                other._gorahn,
                other._prorahn,
            ),
        )

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, DniDatetime):
            return self._cmp(other) == 0
        if not isinstance(other, DniDate):
            return NotImplemented
        return False

    def __le__(self, other: Any) -> bool:
        if isinstance(other, DniDatetime):
            return self._cmp(other) <= 0
        if not isinstance(other, DniDate):
            return NotImplemented
        return NotImplemented

    def __lt__(self, other: Any) -> bool:
        if isinstance(other, DniDatetime):
            return self._cmp(other) < 0
        if not isinstance(other, DniDate):
            return NotImplemented
        return NotImplemented

    def __ge__(self, other: Any) -> bool:
        if isinstance(other, DniDatetime):
            return self._cmp(other) >= 0
        if not isinstance(other, DniDate):
            return NotImplemented
        return NotImplemented

    def __gt__(self, other: Any) -> bool:
        if isinstance(other, DniDatetime):
            return self._cmp(other) > 0
        if not isinstance(other, DniDate):
            return NotImplemented
        return NotImplemented

    def date(self) -> DniDate:
        """Return the date part of the DniDatetime."""
        return DniDate(self._hahr, self._vailee, self._yahr)

    def time(self) -> DniTime:
        """Return the time part of the DniDatetime."""
        return DniTime(self._gahrtahvo, self._tahvo, self._gorahn, self._prorahn)

    def timestamp(self) -> int:
        """Return total number of prorahntee since the D'ni epoch."""
        return (self - DNI_EPOCH_DDT).total_prorahntee() #type: ignore

    def timetuple(self) -> DniDateTimeTuple:
        """Return the DniDateTime in the form of a timetuple."""
        return (
            self._hahr,
            self._vailee,
            self._yahr,
            self._gahrtahvo,
            self._tahvo,
            self._gorahn,
            self._prorahn
        )

    def to_ordinal(self) -> int:
        """Return a D'ni ordinal for the hahr, vailee, and yahr."""
        return hvy2ord(self._hahr, self._vailee, self._yahr)

    @classmethod
    def combine(cls, date_obj: DniDate, time_obj: DniTime) -> "DniDatetime":
        """Construct a D'ni datetime from a given DniDate and DniTime."""
        # pylint: disable=protected-access
        if not isinstance(date_obj, DniDate):
            raise TypeError("date_obj argument must be a DniDate instance.")
        if not isinstance(time_obj, DniTime):
            raise TypeError("time_obj argument must be a DniTime instance.")
        return cls(
            date_obj._hahr,
            date_obj._vailee,
            date_obj._yahr,
            time_obj._gahrtahvo,
            time_obj._tahvo,
            time_obj._gorahn,
            time_obj._prorahn,
        )

    def to_earth(self, tz: Optional[timezone] = timezone.utc) -> datetime:
        """Create an earth datetime from a DniDatetime."""
        return dni_to_earth(self.timetuple()).astimezone(tz)

    @classmethod
    def from_earth(cls, date_time: datetime) -> "DniDatetime":
        """Create a DniDatetime from an earth datetime."""
        ddt = earth_to_dni(date_time)
        return cls(*ddt)


DniDatetime.min = DniDatetime(7654, 1, 1)
DniDatetime.max = DniDatetime(17655, 8, 1, 4, 24, 24, 24)
DniDatetime.resolution = DniTimedelta(prorahntee=1)
DNI_EPOCH_DDT = DniDatetime(DNI_EPOCH_HAHR, 1, 1, 0, 0, 0, 0)
