"""The DniTimedelta class."""

from datetime import timedelta
from math import modf
from typing import Any, Tuple, TYPE_CHECKING, Union

from .constants import DNI_UNITS, MAX_DNI_DELTA_YAHRTEE, MS_PER_PRORAHN
from .utils import cmp, divide_and_round


class DniTimedelta:
    """A D'ni timedelta, modelled after the Python datetime.timedelta class."""

    __slots__ = "_yahrtee", "_pahrtahvotee", "_prorahntee"

    if TYPE_CHECKING:
        _yahrtee: int
        _pahrtahvotee: int
        _prorahntee: int

        # Class properties
        min: "DniTimedelta"
        max: "DniTimedelta"
        resolution: "DniTimedelta"

    def __new__(
        cls,
        yahrtee: Union[float, int] = 0,
        pahrtahvotee: Union[float, int] = 0,
        prorahntee: Union[float, int] = 0,
        gorahntee: Union[float, int] = 0,
        tahvotee: Union[float, int] = 0,
        gahrtahvotee: Union[float, int] = 0,
        vaileetee: Union[float, int] = 0,
    ) -> "DniTimedelta":
        # Final values are integer.
        yahr: int = 0
        pahr: int = 0
        pro: int = 0

        # Normalize
        yahrtee += vaileetee * DNI_UNITS["yahr"]["max"]
        pahrtahvotee += gahrtahvotee * 5
        prorahntee += (gorahntee * DNI_UNITS["prorahn"]["max"]) + (
            tahvotee * DNI_UNITS["gorahn"]["max"] * DNI_UNITS["prorahn"]["max"]
        )

        # Remove fractions: yahrtee
        if isinstance(yahrtee, float):
            yahr_frac, yahrtee = modf(yahrtee)
            yahr_pahr_frac, yahr_pahr_whole = modf(yahr_frac * 25.0)
            assert yahr_pahr_whole == int(yahr_pahr_whole)
            pahr = int(yahr_pahr_whole)
            assert yahrtee == int(yahrtee)
            yahr = int(yahrtee)
        else:
            yahr_pahr_frac = 0.0
            yahr = yahrtee
        assert isinstance(yahr_pahr_frac, float)
        assert abs(yahr_pahr_frac) <= 1.0
        assert isinstance(yahr, int)
        assert abs(pahr) <= 25

        # Remove fractions: pahrtahvotee
        if isinstance(pahrtahvotee, float):
            pahr_frac, pahrtahvotee = modf(pahrtahvotee)
            assert pahrtahvotee == int(pahrtahvotee)
            pahrtahvotee = int(pahrtahvotee)
            pahr_frac += yahr_pahr_frac
            assert abs(pahr_frac) <= 2.0
        else:
            pahr_frac = yahr_pahr_frac
        assert isinstance(pahr_frac, float)
        assert abs(pahr_frac) <= 2.0

        assert isinstance(pahrtahvotee, int)
        yahrtee, pahrtahvotee = divmod(pahrtahvotee, 25)
        yahr += yahrtee
        pahr += int(pahrtahvotee)
        assert isinstance(pahr, int)
        assert abs(pahr) <= 2 * 25

        # Remove fractions: prorahntee
        pro_double = pahr_frac * 3125.0
        assert abs(pro_double) < 2 * 3125.0

        if isinstance(prorahntee, float):
            prorahntee = round(prorahntee + pro_double)
            pahrtahvotee, prorahntee = divmod(prorahntee, 3125)
            yahrtee, pahrtahvotee = divmod(pahrtahvotee, 25)
            yahr += yahrtee
            pahr += pahrtahvotee
        else:
            prorahntee = int(prorahntee)
            pahrtahvotee, prorahntee = divmod(prorahntee, 3125)
            yahrtee, pahrtahvotee = divmod(pahrtahvotee, 25)
            yahr += yahrtee
            pahr += pahrtahvotee
            prorahntee = round(prorahntee + pro_double)
        assert isinstance(pahr, int)
        assert isinstance(prorahntee, int)
        assert abs(pahr) <= 3 * 25
        assert abs(prorahntee) < 3 * 3125.0

        # Last bit of normalization
        pahrtahvotee, pro = divmod(prorahntee, 3125)
        pahr += pahrtahvotee
        yahrtee, pahr = divmod(pahr, 25)
        yahr += yahrtee

        assert isinstance(yahr, int)
        assert isinstance(pahr, int) and 0 <= pahr < 25
        assert isinstance(pro, int) and 0 <= pro < 3125

        if abs(yahr) > MAX_DNI_DELTA_YAHRTEE:
            raise OverflowError(f"DniTimedelta # of yahrtee is too large: {yahr}")

        self = object.__new__(cls)
        self._yahrtee = yahr
        self._pahrtahvotee = pahr
        self._prorahntee = pro
        return self

    def __repr__(self) -> str:
        args = []
        if self._yahrtee:
            args.append(f"yahrtee={self._yahrtee}")
        if self._pahrtahvotee:
            args.append(f"pahrtahvotee={self._pahrtahvotee}")
        if self._prorahntee:
            args.append(f"prorahntee={self._prorahntee}")
        if not args:
            args.append("0")
        return f"{self.__class__.__qualname__}" f"({', '.join(args)})"

    def __str__(self) -> str:
        prorahn = self._pahrtahvotee * 3125 + self._prorahntee
        go, pro = divmod(prorahn, 25)
        tahvo, go = divmod(go, 25)
        pahr, tahvo = divmod(tahvo, 5)
        output = f"{pahr}:{tahvo:02}:{go:02}:{pro:02}"
        if self._yahrtee:
            plural = "tee" if self._yahrtee != 1 else ""
            output = f"{self._yahrtee} yahr{plural}, " + output
        return output

    def total_prorahntee(self) -> int:
        """Total number of prorahntee in the duration."""
        return (self._yahrtee * 78125) + (self._pahrtahvotee * 3125) + self._prorahntee

    # Read-only field accessors
    @property
    def yahrtee(self) -> int:
        """yahrtee"""
        return self._yahrtee

    @property
    def pahrtahvotee(self) -> int:
        """pahrtahvotee"""
        return self._pahrtahvotee

    @property
    def prorahntee(self) -> int:
        """prorahntee"""
        return self._prorahntee

    # Arithmetic functions
    def __add__(self, other: Any) -> "DniTimedelta":
        if isinstance(other, (timedelta, DniTimedelta)):
            if isinstance(other, timedelta):
                other = self.from_timedelta(other)
            return DniTimedelta(
                self._yahrtee + other._yahrtee,
                self._pahrtahvotee + other._pahrtahvotee,
                self._prorahntee + other._prorahntee,
            )
        return NotImplemented

    __radd__ = __add__

    def __sub__(self, other: Any) -> "DniTimedelta":
        if isinstance(other, (timedelta, DniTimedelta)):
            if isinstance(other, timedelta):
                other = self.from_timedelta(other)
            return DniTimedelta(
                self._yahrtee - other._yahrtee,
                self._pahrtahvotee - other._pahrtahvotee,
                self._prorahntee - other._prorahntee,
            )
        return NotImplemented

    def __rsub__(self, other: Any) -> "DniTimedelta":
        if isinstance(other, (timedelta, DniTimedelta)):
            if isinstance(other, timedelta):
                other = self.from_timedelta(other)
            return -self + other
        return NotImplemented

    def __neg__(self) -> "DniTimedelta":
        return DniTimedelta(-self._yahrtee, -self._pahrtahvotee, -self._prorahntee)

    def __pos__(self) -> "DniTimedelta":
        return self

    def __abs__(self) -> "DniTimedelta":
        if self._yahrtee < 0:
            return -self
        return self

    def __mul__(self, other: Any) -> "DniTimedelta":
        if isinstance(other, int):
            return DniTimedelta(
                self._yahrtee * other,
                self._pahrtahvotee * other,
                self._prorahntee * other,
            )
        if isinstance(other, float):
            pro = self.total_prorahntee()
            numer, denom = other.as_integer_ratio()
            return DniTimedelta(0, 0, divide_and_round(pro * numer, denom))
        return NotImplemented

    __rmul__ = __mul__

    def __floordiv__(self, other: Any) -> Union["DniTimedelta", int]:
        if not isinstance(other, (int, DniTimedelta)):
            return NotImplemented
        pro = self.total_prorahntee()
        if isinstance(other, DniTimedelta):
            return pro // other.total_prorahntee()
        if isinstance(other, int):
            return DniTimedelta(0, 0, pro // other)

    def __truediv__(self, other: Any) -> Union["DniTimedelta", float]:
        if not isinstance(other, (float, int, DniTimedelta)):
            return NotImplemented
        pro = self.total_prorahntee()
        if isinstance(other, DniTimedelta):
            return pro / other.total_prorahntee()
        if isinstance(other, float):
            numer, denom = other.as_integer_ratio()
            return DniTimedelta(0, 0, divide_and_round(denom * pro, numer))
        if isinstance(other, int):
            return DniTimedelta(0, 0, divide_and_round(pro, other))

    def __mod__(self, other: Any) -> "DniTimedelta":
        if isinstance(other, DniTimedelta):
            rem = self.total_prorahntee() % other.total_prorahntee()
            return DniTimedelta(0, 0, rem)
        return NotImplemented

    def __divmod__(self, other: Any) -> Tuple[int, "DniTimedelta"]:
        if isinstance(other, DniTimedelta):
            quot, rem = divmod(self.total_prorahntee(), other.total_prorahntee())
            return quot, DniTimedelta(0, 0, rem)
        return NotImplemented

    # Rich comparisons
    def _cmp(self, other: Any) -> int:
        """Helper function for rich comparison of DniTimedelta objects."""
        assert isinstance(other, DniTimedelta)
        return cmp(self.total_prorahntee(), other.total_prorahntee())

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, DniTimedelta):
            return self._cmp(other) == 0
        return NotImplemented

    def __le__(self, other: Any) -> bool:
        if isinstance(other, DniTimedelta):
            return self._cmp(other) <= 0
        return NotImplemented

    def __lt__(self, other: Any) -> bool:
        if isinstance(other, DniTimedelta):
            return self._cmp(other) < 0
        return NotImplemented

    def __ge__(self, other: Any) -> bool:
        if isinstance(other, DniTimedelta):
            return self._cmp(other) >= 0
        return NotImplemented

    def __gt__(self, other: Any) -> bool:
        if isinstance(other, DniTimedelta):
            return self._cmp(other) > 0
        return NotImplemented

    def __bool__(self) -> bool:
        return self._yahrtee != 0 or self._pahrtahvotee != 0 or self._prorahntee != 0

    @classmethod
    def from_timedelta(cls, delta: timedelta) -> "DniTimedelta":
        """Converts a normal timedelta to a D'Ni timedelta."""
        prorahntee = delta.total_seconds() * 1000 / MS_PER_PRORAHN
        return cls(prorahntee=prorahntee)

    def to_timedelta(self) -> timedelta:
        """Converts the D'Ni timedelta to a normal timedelta."""
        milliseconds = self.total_prorahntee() * MS_PER_PRORAHN
        try:
            delta = timedelta(milliseconds=milliseconds)
            return delta
        except OverflowError as exc:
            # This is needed for precision issues in conversion.
            if self >= self.max:
                return timedelta.max
            if self <= self.min:
                return timedelta.min
            raise exc


# Converted from timedelta.min
DniTimedelta.min = DniTimedelta(yahrtee=-793993715, pahrtahvotee=17, prorahntee=768)

# Converted from timedelta.max
DniTimedelta.max = DniTimedelta(yahrtee=793993715, pahrtahvotee=2, prorahntee=1887)

# Smallest measurement of change.
DniTimedelta.resolution = DniTimedelta(prorahntee=1)
