"""Helper functions for all classes/functions of the module."""

# pylint: disable=line-too-long

from datetime import datetime, timedelta, timezone
from typing import Any, Optional, Tuple, Union

from .constants import DNI_UNITS, EARTH_EPOCH, LEAP_MS_FROM_EPOCH


def cmp(x: Any, y: Any) -> int:
    """Compare objects and return:

    x < y  -> -1
    x == y ->  0
    x > y  ->  1
    """
    return 0 if x == y else 1 if x > y else -1


def divide_and_round(a: Union[float, int], b: Union[float, int]) -> int:
    """Divide a by b and round result to the nearest integer. When the ratio
    is exactly half-way between two integers, the even integer is returned.
    """
    quot, rem = divmod(a, b)
    rem *= 2
    greater_than_half = rem > b if b > 0 else rem < b
    if greater_than_half or rem == b and quot % 2 == 1:
        quot += 1

    return int(quot)


def check_dni_date_fields(
    hahr: int = DNI_UNITS["hahr"]["min"],
    vailee: int = DNI_UNITS["vailee"]["min"],
    yahr: int = DNI_UNITS["yahr"]["min"],
) -> Tuple[int, int, int]:
    """Verifies the D'Ni date fields are within proper boundaries."""
    if not DNI_UNITS["hahr"]["min"] <= hahr <= DNI_UNITS["hahr"]["max"]:
        raise ValueError(
            f"Hahr must be in {DNI_UNITS['hahr']['min']}..{DNI_UNITS['hahr']['max']}",
            hahr,
        )
    if not DNI_UNITS["vailee"]["min"] <= vailee <= DNI_UNITS["vailee"]["max"]:
        raise ValueError(
            f"Vailee must be in {DNI_UNITS['vailee']['min']}..{DNI_UNITS['vailee']['max']}",
            vailee,
        )
    if not DNI_UNITS["yahr"]["min"] <= yahr <= DNI_UNITS["yahr"]["max"]:
        raise ValueError(
            f"Yahr must be in {DNI_UNITS['yahr']['min']}..{DNI_UNITS['yahr']['max']}",
            yahr,
        )
    return hahr, vailee, yahr


def check_dni_time_fields(
    gahrtahvo: int = DNI_UNITS["gahrtahvo"]["min"],
    tahvo: int = DNI_UNITS["tahvo"]["min"],
    gorahn: int = DNI_UNITS["gorahn"]["min"],
    prorahn: int = DNI_UNITS["prorahn"]["min"],
) -> Tuple[int, int, int, int]:
    """Verifies the D'Ni time fields are within proper boundaries."""
    if (
        not DNI_UNITS["gahrtahvo"]["min"]
        <= gahrtahvo
        <= DNI_UNITS["gahrtahvo"]["max"] - 1
    ):
        raise ValueError(
            f"Gahrtahvo must be in {DNI_UNITS['gahrtahvo']['min']}..{DNI_UNITS['gahrtahvo']['max'] - 1}",
            gahrtahvo,
        )
    if not DNI_UNITS["tahvo"]["min"] <= tahvo <= DNI_UNITS["tahvo"]["max"] - 1:
        raise ValueError(
            f"Tahvo must be in {DNI_UNITS['tahvo']['min']}..{DNI_UNITS['tahvo']['max'] - 1}",
            tahvo,
        )
    if (
        not DNI_UNITS["gorahn"]["min"]
        <= gorahn
        <= DNI_UNITS["gorahn"]["max"] - 1
    ):
        raise ValueError(
            f"Gorahn must be in {DNI_UNITS['gorahn']['min']}..{DNI_UNITS['gorahn']['max'] - 1}",
            gorahn,
        )
    if (
        not DNI_UNITS["prorahn"]["min"]
        <= prorahn
        <= DNI_UNITS["prorahn"]["max"] - 1
    ):
        raise ValueError(
            f"Prorahn must be in {DNI_UNITS['prorahn']['min']}..{DNI_UNITS['prorahn']['max'] - 1}",
            prorahn,
        )
    return gahrtahvo, tahvo, gorahn, prorahn


def add_leap_seconds(timestamp: int) -> int:
    """Adds leap seconds to a timestamp."""
    leap_seconds = 0
    for leap in LEAP_MS_FROM_EPOCH:
        if timestamp >= leap:
            leap_seconds += 1

    # Adjust for the 10 leap seconds that started in 1972.
    if leap_seconds > 0:
        leap_seconds += 9

    return timestamp + (leap_seconds * 1000)


def get_ms_from_epoch(date_time: Optional[datetime] = None) -> int:
    """
    Returns the amount of milliseconds since the UNIX epoch (1-1-1970).
    """
    # Thanks to the examples from here:
    # https://stackoverflow.com/questions/5395872/
    if not date_time:
        date = datetime.now(timezone.utc)
    else:
        date = date_time.replace(tzinfo=timezone.utc)
    return ((date - EARTH_EPOCH) // timedelta(microseconds=1)) // 1000


def get_adj_ms_from_epoch(date_time: Optional[datetime] = None) -> int:
    """
    Returns the amount of milliseconds since the UNIX epoch (1-1-1970),
    but also accounting for leap seconds.
    """
    return add_leap_seconds(get_ms_from_epoch(date_time))


def is_timezone_aware(dt: datetime) -> bool:
    """Returns whether a datetime object has timezone info or not."""
    return dt.tzinfo is not None and dt.tzinfo.utcoffset(dt) is not None
